/*****************************************************************************
 * PONG - demo DirectFB software for STK1000 BSP
 *
 * See README for program description and instructions.
 * 
 * Atmel Corporation: http://www.atmel.com
 *   Support email: avr32@atmel.com
 *
 *****************************************************************************/

/* Copyright (c) 2006, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <directfb.h>

#define PONG_BG "/etc/stk1000splash.jpg"

/* macro for a safe call to DirectFB functions */
#define DFBCHECK(x...) \
     {                                                                \
          err = x;                                                    \
          if (err != DFB_OK) {                                        \
               fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ ); \
               DirectFBErrorFatal( #x, err );                         \
          }                                                           \
     }

int pong_play();
void pong_update_score(int p1_score, int p2_score, IDirectFBWindow *window);
void pong_setbg(char *image);

IDirectFB              *dfb	= NULL;
IDirectFBFont          *font	= NULL;
IDirectFBDisplayLayer  *layer;
DFBDisplayLayerConfig  layer_config;

int gpio;

/*! Main
 *
 *  \param argc number of arguments
 *  \param argv array of arguments
 *
 *  \retval exit status
 */
int main(int argc, char *argv[])
{
	int err;

	gpio = open("/dev/gpio0", O_RDONLY|O_NONBLOCK);

	if (gpio < 0) {
		int retval = errno;
		printf("could not open /dev/gpio0, exiting with errno %d", retval);
		return retval;
	}

	DFBGraphicsDeviceDescription gdesc;

	DFBCHECK(DirectFBInit(&argc, &argv));
	DFBCHECK(DirectFBCreate(&dfb));

	dfb->GetDeviceDescription(dfb, &gdesc);

	DFBCHECK(dfb->GetDisplayLayer(dfb, DLID_PRIMARY, &layer));

	layer->SetCooperativeLevel(layer, DLSCL_ADMINISTRATIVE);

	if (!((gdesc.blitting_flags & DSBLIT_BLEND_ALPHACHANNEL) &&
				(gdesc.blitting_flags & DSBLIT_BLEND_COLORALPHA )))
	{
		layer_config.flags = DLCONF_BUFFERMODE | DLCONF_SURFACE_CAPS;
		layer_config.buffermode = DLBM_BACKVIDEO;
		layer_config.surface_caps = DSCAPS_VIDEOONLY;
		layer->SetConfiguration(layer, &layer_config);
	}

	layer->GetConfiguration(layer, &layer_config);

	{
		DFBFontDescription font_desc;
		font_desc.flags = DFDESC_HEIGHT;
		font_desc.height = 16;

		DFBCHECK(dfb->CreateFont(dfb, FONT, &font_desc, &font));
	}

	pong_setbg(PONG_BG);

	do {
		err = pong_play();
	} while (err > 0);

	/* Set black surface */
	layer->SetBackgroundMode(layer, DLBM_COLOR);
	layer->SetBackgroundColor(layer, 0x00, 0x00, 0x00, 0x00);
	layer->SetOpacity(layer, 0xff);

	/* clean up */
	if (font)
		font->Release(font);
	if (layer)
		layer->Release(layer);
	if (dfb)
		dfb->Release(dfb);
	if (gpio)
		close(gpio);

	return 0;
}

/*! Play pong style game
 *
 *  \return if the game should be restarted or application exit
 *    \retval 0 on quit signal
 *    \retval 1 on restart signal
 */
int pong_play()
{
	IDirectFBWindow		*player_w1;
	IDirectFBWindow		*player_w2;
	IDirectFBWindow		*ball_w;
	IDirectFBWindow		*score_w;
	IDirectFBSurface	*player_surface1;
	IDirectFBSurface	*player_surface2;
	IDirectFBSurface	*ball_surface;
	IDirectFBSurface	*score_surface;

	int err;
	uint32_t buttons;

	int pos_x_p1	= 10;
	int pos_x_p2	= 300;
	int p_width	= 10;
	int p_height	= 70;
	int ball_width	= 10;
	int ball_height	= 10;

	int p1_score	= 0;
	int p2_score	= 0;
	int pos_y_p1	= 80;
	int pos_y_p2	= 80;
	int pos_x_ball	= 150;
	int pos_y_ball	= 110;

	int retval		= 0;
	short game_on		= 1;
	short running		= 0;
	short dir_ball_x	= 1;
	short dir_ball_y	= 1;
	short ball_speed_x	= 2;
	short ball_speed_y	= 0;

	/* Scorefield setup */
	{
		DFBWindowDescription desc;

		desc.flags  = (DWDESC_POSX | DWDESC_POSY |
				DWDESC_WIDTH | DWDESC_HEIGHT | DWDESC_CAPS);
		desc.posx   = 110;
		desc.posy   = 10;
		desc.width  = 100;
		desc.height = 20;
		desc.caps   = DWCAPS_ALPHACHANNEL;

		DFBCHECK(layer->CreateWindow(layer, &desc, &score_w));
		DFBCHECK(score_w->GetSurface(score_w, &score_surface));
		DFBCHECK(score_surface->SetFont(score_surface, font));

		score_w->DisableEvents(score_w, DWET_ALL);
		score_w->SetOpacity(score_w, 0x80);

		score_surface->SetColor(score_surface,
				0x80, 0x1e, 0xff, 0x00);
		score_surface->FillRectangle(score_surface,
				1, 1,
				desc.width-2, desc.height-2);

		score_surface->Flip(score_surface, NULL, 0);

		score_w->RequestFocus(score_w);
		score_w->RaiseToTop(score_w);
		pong_update_score(p1_score, p2_score, score_w);
	}

	/* Player 1 setup */
	{
		DFBWindowDescription desc;

		desc.flags = (DWDESC_POSX | DWDESC_POSY |
				DWDESC_WIDTH | DWDESC_HEIGHT);

		desc.posx   = pos_x_p1;;
		desc.posy   = pos_y_p1;
		desc.width  = p_width;
		desc.height = p_height;

		DFBCHECK(layer->CreateWindow(layer, &desc, &player_w1));

		player_w1->DisableEvents(player_w1, DWET_ALL);
		player_w1->GetSurface(player_w1, &player_surface1);

		player_w1->SetOpacity(player_w1, 0xff);
		player_surface1->SetColor(player_surface1,
				0x80, 0x40, 0x40, 0x00);
		player_surface1->DrawRectangle(player_surface1,
				0, 0, desc.width, desc.height);
		player_surface1->SetColor(player_surface1,
				0xff, 0x00, 0x00, 0x00);
		player_surface1->FillRectangle(player_surface1,
				1, 1,
				desc.width-2, desc.height-2);
		player_surface1->Flip(player_surface1, NULL, 0);
		player_w1->RequestFocus(player_w1);
		player_w1->RaiseToTop(player_w1);
	}

	/* Player 2 setup */
	{
		DFBWindowDescription desc;

		desc.flags = (DWDESC_POSX | DWDESC_POSY |
				DWDESC_WIDTH | DWDESC_HEIGHT);

		desc.posx   = pos_x_p2;;
		desc.posy   = pos_y_p2;
		desc.width  = p_width;
		desc.height = p_height;

		DFBCHECK(layer->CreateWindow(layer, &desc, &player_w2));
		player_w2->DisableEvents(player_w2, DWET_ALL);

		player_w2->GetSurface(player_w2, &player_surface2);
		player_w2->SetOpacity(player_w2, 0xff);
		player_surface2->SetColor(player_surface2,
				0x40, 0x80, 0x40, 0xff);
		player_surface2->DrawRectangle(player_surface2,
				0, 0, desc.width, desc.height);
		player_surface2->SetColor(player_surface2,
				0x00, 0xff, 0x00, 0x00);
		player_surface2->FillRectangle(player_surface2,
				1, 1,
				desc.width-2, desc.height-2);

		player_surface2->Flip(player_surface2, NULL, 0);

		player_w2->RequestFocus(player_w2);
		player_w2->RaiseToTop(player_w2);
	}

	/* Ball setup */
	{
		DFBWindowDescription desc;

		desc.flags = (DWDESC_POSX | DWDESC_POSY |
				DWDESC_WIDTH | DWDESC_HEIGHT);

		desc.posx   = pos_x_ball;
		desc.posy   = pos_y_ball;
		desc.width  = ball_width;
		desc.height = ball_height;

		DFBCHECK(layer->CreateWindow(layer, &desc, &ball_w));
		ball_w->DisableEvents(ball_w, DWET_ALL);
		ball_w->GetSurface(ball_w, &ball_surface);
		ball_w->SetOpacity(ball_w, 0xff);

		ball_surface->SetColor(ball_surface,
				0xf0, 0xa0, 0x50, 0xff);
		ball_surface->FillRectangle(ball_surface,
				1, 1,
				desc.width, desc.height-2);

		ball_surface->SetColor(ball_surface,
				0x20, 0x20, 0x20, 0xff);
		ball_surface->DrawRectangle(ball_surface,
				0, 0, desc.width, desc.height);

		ball_surface->Flip(ball_surface, NULL, 0);
		ball_w->RequestFocus(ball_w);
		ball_w->RaiseToTop(ball_w);
	}

	while (game_on) {
		read(gpio, &buttons, sizeof(buttons));

		player_w1->GetPosition(player_w1, &pos_x_p1, &pos_y_p1);
		player_w2->GetPosition(player_w2, &pos_x_p2, &pos_y_p2);

		if (~buttons & 0x40) /* switch up player 1 */
			if (pos_y_p1 < 170)
				player_w1->Move(player_w1, 0, 2);

		if (~buttons & 0x80) /* switch down player 1 */
			if (pos_y_p1 > 5)
				player_w1->Move(player_w1, 0, -2);

		if (~buttons & 0x01) /* switch up player 2 */
			if (pos_y_p2 < 170)
				player_w2->Move(player_w2, 0, 2);

		if (~buttons & 0x02) /* switch down player 2 */
			if (pos_y_p2 > 5)
				player_w2->Move(player_w2, 0, -2);

		/* Wait for it, wait, wait ... (delay to make it playable) */
		usleep(10000);
		/* GO! */

		if (running) {
			/* Ball is to the left, check if it hit the player bar */
			if (pos_x_ball <= 20) {
				/* Player bar was another place */
				if (pos_y_ball > (pos_y_p1 + p_height)
						|| pos_y_ball < (pos_y_p1 - ball_height)) {
					p1_score++;
					running=0;
					pong_update_score(p1_score, p2_score, score_w);
					ball_w->MoveTo(ball_w , 150, 110);
					ball_speed_x = 2;
					ball_speed_y = 0;
				/* Player bar hit the ball, send it out */
				} else {
					if (pos_y_ball > (pos_y_p1 + 3 * p_height/4)){
						ball_speed_y++;
						ball_speed_x=1;
					}
					else if (pos_y_ball < (pos_y_p1 + p_height/4)){
						ball_speed_y++;
						ball_speed_x=1;
					} else {
						ball_speed_y=1;
						ball_speed_x++;
					}
				}
			}
			/* Ball is to the right, check if it hit the player bar */
			else if (pos_x_ball >= (300 - ball_height)) {
				/* Player bar was another place */
				if (pos_y_ball > (pos_y_p2+p_height)
						|| pos_y_ball < (pos_y_p2 - ball_height))
				{
					p2_score++;
					running=0;
					pong_update_score(p1_score, p2_score, score_w);
					ball_w->MoveTo(ball_w , 150, 110);
					ball_speed_x = 2;
					ball_speed_y = 0;
				/* Player bar hit the ball, send it out */
				} else {
					if (pos_y_ball > (pos_y_p2 + 3 * p_height/4)){
						ball_speed_y++;
						ball_speed_x=1;
					}
					else if (pos_y_ball < (pos_y_p2 + p_height/4)){
						ball_speed_y++;
						ball_speed_x=1;
					} else {
						ball_speed_y=1;
						ball_speed_x++;
					}
				}
			}

			/* Check if ball has struck the top or bottom and
			 * change the direction
			 */
			if ((pos_y_ball > 10) && (pos_y_ball < (240 - ball_height))) {
				dir_ball_y = dir_ball_y;
			} else {
				dir_ball_y = -dir_ball_y;
			}

			/* Check if ball has struck the left or right side and
			 * change the direction
			 */
			if (pos_x_ball > 20 && pos_x_ball < (300 - ball_height)) {
				dir_ball_x = dir_ball_x;
			} else {
				dir_ball_x = -dir_ball_x;
			}

			ball_w->Move(ball_w, dir_ball_x * ball_speed_x,
					dir_ball_y * ball_speed_y);
			ball_w->GetPosition(ball_w, &pos_x_ball, &pos_y_ball);

			/* check for quit */
			if (~buttons & 0x10) {
				retval = 0;
				game_on = 0;
				running = 0;
			}
			/* check for game end */
			else if (~buttons & 0x08) {
				retval = 1;
				game_on = 0;
				running = 0;
			}
		} else {
			/* check for quit */
			if (~buttons & 0x10) {
				retval = 0;
				game_on = 0;
			}
			/* check for game end */
			else if (~buttons & 0x08) {
				retval = 1;
				game_on = 0;
				running = 0;
			}
			/* check for start button */
			else if (~buttons & 0x04) {
				running = 1;
			}
		}
	} /* while game on */

	/* Clean up */
	player_w1->SetOpacity(player_w1, 0x00);
	player_w1->SetOpacity(player_w2, 0x00);
	player_w1->SetOpacity(ball_w, 0x00);
	player_w1->SetOpacity(score_w, 0x00);

	player_surface1->Release(player_surface1);
	player_surface2->Release(player_surface2);
	ball_surface->Release(ball_surface);
	score_surface->Release(score_surface);

	player_w1->Release(player_w1);
	player_w2->Release(player_w2);
	ball_w->Release(ball_w);
	score_w->Release(score_w);

	return retval;
}

/*! Updates the pong highscore
 *
 *  \param p1_score score number for player 1
 *  \param p2_score score number for player 2
 *  \param window pointer to DirectFB window to draw in
 */
void pong_update_score(int p1_score, int p2_score, IDirectFBWindow *window)
{
	IDirectFBSurface       *primary;
	int window_height=0;
	int window_width=0;
	int fontheight;
	int err;
	char text[20];

	sprintf(text, "P1:%d P2:%d", p1_score, p2_score);

	DFBCHECK(window->GetSurface(window, &primary));
	DFBCHECK(primary->SetColor(primary, 0x0, 0x0, 0x0, 0xff));
	DFBCHECK(primary->GetSize(primary, &window_width, &window_height));
	DFBCHECK(primary->FillRectangle(primary, 0, 0, window_width, window_height));

	font->GetHeight(font, &fontheight);
	DFBCHECK(primary->SetColor(primary, 0x8f, 0xff, 0x20, 0xff));
	DFBCHECK(primary->DrawString(primary, text, -1, window_width/2,
				window_height - fontheight/4, DSTF_CENTER));

	primary->Flip(primary, NULL, 0);
	primary->Release(primary);
	window->RequestFocus(window);
	window->RaiseToTop(window);
	sleep(1);
}

/*! Function to set the background image
 *
 *  \param image path and name of image to load
 */
void pong_setbg(char *image)
{
	int err;

	IDirectFBImageProvider *provider;
	IDirectFBSurface       *bgsurface;
	DFBSurfaceDescription  desc;

	desc.flags  = DSDESC_WIDTH | DSDESC_HEIGHT;
	desc.width  = layer_config.width;
	desc.height = layer_config.height;

	DFBCHECK(dfb->CreateImageProvider(dfb, image, &provider));
	DFBCHECK(dfb->CreateSurface(dfb, &desc, &bgsurface));

	DFBCHECK(provider->RenderTo(provider, bgsurface, NULL));

	DFBCHECK(layer->SetBackgroundImage(layer, bgsurface));
	DFBCHECK(layer->SetBackgroundMode(layer, DLBM_IMAGE));

	DFBCHECK(bgsurface->Release(bgsurface));
	DFBCHECK(provider->Release(provider));
}
